// ----------------------------------------------------------------------------
// CERTI - HLA Run Time Infrastructure
// Copyright (C) 2007-2010 Andrej Pancik, Yannick Bisiaux, Ronan Bossard, Samuel Reese
//
// This program is free software ; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation ; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY ; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this program ; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
// ----------------------------------------------------------------------------
package certi.communication;

/**
 * <p><code>CertiMessage</code> represents one of the messages which are exchanged between the federate and the RTIA.
 * It contains the information that are used in all messages (the <code>CertiMessageType</code> of the message, its <code>CertiExceptionType</code> and its <code>FederationTime</code>)</p>
 *
 * @author apancik@gmail.com
 * @author <a href = "mailto:yannick.bisiaux@supaero.fr">Yannick Bisiaux</a>, <a href = "mailto:ronan.bossard@supaero.fr">Ronan Bossard</a>, <a href = "mailto:samuel.reese@supaero.fr">Samuel Reese</a>
 * @version 1.0
 */
public class CertiMessage {

    private CertiMessageType type;
    protected CertiExceptionType exception;
    private double federationTime;
    private String reason;

    /**Creates a new instance of <code>CertiMessage</code>. The default initialization set CertiMessageType to NOT_USED, CertiExceptionType to e_NO_EXCEPTION, FederationTime to 0.
     *
     */
    public CertiMessage() {
        this.type = CertiMessageType.NOT_USED;
        this.exception = CertiExceptionType.NO_EXCEPTION;
        this.federationTime = 0;
    }

    /**Creates a new instance of <code>CertiMessage</code>.
     *
     * @param type the <code>CertiMessageType</code> to set
     */
    public CertiMessage(CertiMessageType type) {
        this.type = type;
        this.exception = CertiExceptionType.NO_EXCEPTION;
        this.federationTime = 0;
    }

    /**Creates a new instance of <code>CertiMessage</code>.
     *
     * @param type the <code>CertiMessageType</code> to set
     * @param exception the <code>CertiExceptionType</code> to set
     * @param federationTime the <code>FederationTime</code> to set
     */
    public CertiMessage(CertiMessageType type, double federationTime) {
        this.type = type;
        this.exception = CertiExceptionType.NO_EXCEPTION;
        this.federationTime = federationTime;
    }

    /**Writes in the <code>MessageBuffer</code> the state of the attributes of the <code>CertiMessage</code>.
     * The order will be the <code>CertiMessageType</code>, the <code>CertiExceptionType</code> and the federation time of the <code>CertiMessage</code>.
     *
     * @param messageBuffer the <code>MessageBuffer</code> in which the method will write the state of the <code>CertiMessage</code>
     */
    public void writeMessage(MessageBuffer messageBuffer) {
        //writeHeader
        messageBuffer.reset();
        messageBuffer.write(this.type.getType());
        messageBuffer.write(this.exception.getException());
        messageBuffer.write(this.federationTime);
    }

    /**Gives the correct values to the attributes by reading a <code>MessageBuffer</code> in which are stored those values.
     *
     * @param messageBuffer the <code>MessageBuffer</code> from which the method will read the state of the <code>CertiMessage</code>
     */
    public void readMessage(MessageBuffer messageBuffer) throws CertiException {
        //this.type = CertiMessageType.reverseType.get(messageBuffer.readInt()); //dont need to read type as it is read before instantiating
        this.exception = CertiExceptionType.reverse.get(messageBuffer.readInt());

        if (this.exception != CertiExceptionType.NO_EXCEPTION) {
            // If the message carry an exception, the Body will only contain the
            // exception reason.

            System.out.println("Received message exception " + this.exception.toString());

            this.reason = messageBuffer.readString();

            throw new CertiException(this.getType(),this.exception, this.reason);
        }

        this.federationTime = messageBuffer.readDouble();
    }

    public String getReason() {
        return reason;
    }

    /**<p>Creates a new default instance of a specialized <code>CertiMessage</code>.
     * The kind of specialized class <code>CertiMessage</code> created is linked with the number in input.</p>
     *
     * <p>For instance, <code>CertiMessage.virtualCreate(1)</code> will return a default instance of <code>MSG_Create_Federation_Execution</code>.</p>
     *
     * @param n an <code>integer</code> which represents the <code>CertiMessageType</code> (kind) of <code>CertiMessage</code> you want to create
     * @return a default instance of the specialized <code>CertiMessage</code>
     */
    public static CertiMessage virtualCreate(int n) {
        String messageClass;
        messageClass = conversionType(n);

        try {
            Class c = Class.forName("certi.communication.messages." + messageClass);//We have to add "message.MSG_" to have the package and the suffix of the class name.
            try {
                Object obj = c.newInstance();
                CertiMessage msg = (CertiMessage) obj;
                msg.setType(CertiMessageType.reverseType.get(n));
                return msg;
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            } catch (InstantiationException e) {
                e.printStackTrace();
            }
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }
        return null;
    }

    //This function takes in input the number of a CertiMessageType of CertiMessage. It returns the name of the class we want to call.
    //Ex : conversionType(2) -> Destroy_Federation_Execution
    private static String conversionType(int n) {
        String type;
        String class_called;

        type = CertiMessageType.reverseString.get(n);
        class_called = CertiMessage.correctClassName(type);

        return class_called;
    }

    //This function takes in input the string of a CertiMessageType of CertiMessage. It returns the string with the capital letters in the right place (after each underscore).
    //Ex : correctClassName(DESTROY_FEDERATION_EXECUTION) -> Destroy_Federation_Execution
    private static String correctClassName(String str) {
        StringBuffer buffer;
        boolean capitalizeNext = true; //The first letter will be a capital letter.
        char ch;

        str = str.toLowerCase(); //We begin by set all characters in lower case.
        buffer = new StringBuffer(str.length());

        for (int i = 0; i < str.length(); i++) {
            ch = str.charAt(i);

            //If the character is an underscore, the next character will be a capital letter.
            if (ch == '_') {
                //buffer.append(ch);
                capitalizeNext = true;
            } else if (capitalizeNext) {
                buffer.append(Character.toTitleCase(ch));
                capitalizeNext = false;
            } else {
                buffer.append(ch);
            }
        }
        return buffer.toString();
    }

    /**Returns the <code>CertiMessageType</code> of this <code>CertiMessage</code>.
     * @return the <code>CertiMessageType</code>
     */
    public CertiMessageType getType() {
        return type;
    }

    /**Sets the <code>CertiMessageType</code> of this <code>CertiMessage</code>.
     * @param type the <code>CertiMessageType</code> to set
     */
    public void setType(CertiMessageType type) {
        this.type = type;
    }

    /**Returns the <code>CertiExceptionType</code> of this <code>CertiMessage</code>.
     * @return the <code>CertiExceptionType</code>
     */
    public CertiExceptionType getException() {
        return exception;
    }

    /**Sets the <code>CertiExceptionType</code> of this <code>CertiMessage</code>.
     * @param exception_ the <code>CertiExceptionType</code> to set
     */
    public void setException(CertiExceptionType exception_) {
        this.exception = exception_;
    }

    /**Returns the <code>FederationTime</code> of this <code>CertiMessage</code>.
     * @return the <code>FederationTime</code>
     */
    public double getFederationTime() {
        return federationTime;
    }

    /**Sets the <code>FederationTime</code> of this <code>CertiMessage</code>.
     * @param federationTime the <code>FederationTime</code> to set
     */
    public void setFederationTime(double federationTime) {
        this.federationTime = federationTime;
    }

    /**This function is able to give a <code>String</code> which represents the state of the message.
     * It gives its <code>CertiMessageType</code>, its <code>CertiExceptionType</code> and the federation time.
     * @return the <code>String</code> which represents the object.
     */
    @Override
    public String toString() {
        String str;

        str = this.type + ", " + this.exception + ", federation time=" + this.federationTime;
        return str;
    }
}
