// ----------------------------------------------------------------------------
// CERTI - HLA Run Time Infrastructure
// Copyright (C) 2007-2010 Andrej Pancik, Yannick Bisiaux, Ronan Bossard, Samuel Reese
//
// This program is free software ; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation ; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY ; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this program ; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
// ----------------------------------------------------------------------------
package certi.communication;

import hla.rti.RTIexception;
import java.util.Vector;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;

/**
 * This class is used to process all the primary data type and to store them into a Vector<Byte>. Depending on
 * how the Socket works, the output will change.
 *
 * The JVM is big endian, so all the write methods are big endians.
 * the read methods can handle big and little endians.
 *
 * @author apancik@gmail.com
 * @author <a href = "mailto:yannick.bisiaux@supaero.fr">Yannick Bisiaux</a>, <a href = "mailto:ronan.bossard@supaero.fr">Ronan Bossard</a>, <a href = "mailto:samuel.reese@supaero.fr">Samuel Reese</a>
 * @version 1.0
 */
public class MessageBuffer {
    //Use a Vector of Byte: no problems of array resizing. easily changed in array if necessary thanks to toArray()
    //Beware! use of Byte instead of byte. Has an influence in the send method.

    private Vector<Byte> buff = new Vector<Byte>();
    private Iterator<Byte> ite;	//the Iterator will be useful for the read operations
    private static byte BIG_ENDIAN = 1;//not really useful, but helps to easily exchange the value of BIG and LITTLE
    private byte endianness;

    //the constructor(s) might change depending on how we use the MessageBuffer.
    /**
     *
     */
    public MessageBuffer() {
        //the first five bytes are for the header. If this buffer is used to read instead of write, they'll be over-written
        //in the receive method
        endianness = BIG_ENDIAN;
        reset();
    }

    /**
     * This method set the buffer back in the state just after its creation
     * should be used before re-using the buffer to write.
     * It is already aotomatically done when a message is read from the socket
     */
    public void reset() {
        buff.clear();
    }

    /**
     * This method just copy the byte b at the end of the buffer. It's the most basic writing method.
     * @param b
     */
    public void write(byte b) {
        buff.add(b);
    }

    /**
     * Basic read method.
     * @return just the byte we were looking for
     */
    public byte readByte() throws NoSuchElementException {
        return ite.next();
    }

    /**
     * Write an array of bytes in the buffer. First is the length of the array, stored as a long, then the bytes themselves
     * Even if the length of the array is stored in a long int, java is limited to array with 32 bits int length.
     * @param byteArray
     */
    public void write(byte[] byteArray) {
        //this.write((long) byteArray.length, "longueur du tableau");
        for (byte b : byteArray) {
            buff.add(b);
        }
    }

    /**
     * read an Array of bytes, given its size
     * @param size the number of bytes to be read
     * @return
     */
    public byte[] readBytes(int size) {
        byte[] byteArray = new byte[size];
        for (int i = 0; i <= size - 1; i++) {
            byteArray[i] = this.readByte();
        }
        return byteArray;
    }

    public byte[] readBytesWithSize() {
        int size = this.readInt(); //Treat first integer as length
        byte[] byteArray = new byte[size];
        for (int i = 0; i <= size - 1; i++) {
            byteArray[i] = this.readByte();
        }
        return byteArray;
    }

    /**
     * Write in a big endian way
     * @param s
     *
     */
    public void write(short s) {
        buff.add((byte) (s >>> 8));
        buff.add((byte) s);
    }

    /**
     * The endianness is taken care of.
     * Beware! there are no unsigned integers of any size in java.
     * This methods assume that the short it is trying to read is inside the bounds of a signed short.
     * There are no conversion to an int if it is not the case.
     * @return
     */
    public short readShort() throws NoSuchElementException {
        int s;	// the << just don't work here with the short, according to Eclipse.
        //
        //I assume that the ite.next occur in the "left to right" order
        //no implicit transtyping
        //if the byte we read is <0, the transtyping puts some 1's to fill the int.
        //the 0x00FF mask set them to 0.
        //I could use  if(endianity), but i won't do this before i'm SURE I didn't swapped the value of BIG_ENDIAN and LITTLE_ENDIAN
        if (endianness == BIG_ENDIAN) {
            s = ((int) ite.next() & 0x00FF) << 8 | ((int) ite.next() & 0x00FF);
        } else {
            s = ((int) ite.next() & 0x00FF) | ((int) ite.next() & 0x00FF) << 8;
        }
        return (short) s;
    }

    /**
     * Write in a big endian way
     * @param b
     */
    public void write(int b) {
        for (int i = 3; i >= 0; i--) {
            buff.add((byte) (b >>> i * 8));
        }
        //java is big-endian.
        //the type conversion in byte take the 8 last (low weight) bits
        //So I move the int to the right to take the byte I need until the 4 bytes have been processed.
        //I use >>> instead of >> to be sure all the added bits on the left are 0s.
    }

    /**
     * The endianness is taken care of.
     * Beware! there are no unsigned integers of any size in java.
     * This methods assume that the int it is trying to read is inside the bounds of a signed int.
     * There are no conversion to an long if it is not the case.
     * @return an int
     */
    public int readInt() throws NoSuchElementException {
        int i = 0;
        if (endianness == BIG_ENDIAN) {
            for (int s = 3; s >= 0; s--) {
                i = i | (((int) ite.next()) & 0x000000FF) << 8 * s;
            }
        } else {
            for (int s = 0; s <= 3; s++) {
                i = i | ((int) ite.next() & 0x000000FF) << s * 8;
            }
        }
        return i;
    }

    /**
     * Write in a big endian way
     * @param l
     */
    public void write(long l) {
        int i;
        for (i = 7; i >= 0; i--) {
            buff.add((byte) (l >>> i * 8));
        }
    }

    /**
     * The endianness is taken care of.
     * Beware! there are no unsigned integers of any size in java.
     * This methods assume that the long it is trying to read is inside the bounds of a signed long.
     * There are no conversion to something bigger if it is not the case.
     * @return
     */
    public long readLong() throws NoSuchElementException {
        long l = 0;
        if (endianness == BIG_ENDIAN) {
            for (int i = 7; i >= 0; i--) {
                l = l | ((long) ite.next() & 0x00000000000000FF) << i * 8;
            }
        } else {
            for (int i = 0; i <= 7; i++) {
                l = l | ((long) ite.next() & 0x00000000000000FF) << i * 8;
            }
        }
        return l;
    }

    /**
     * Write in a big endian, IEEE 754 way
     * @param f
     */
    public void write(float f) throws NoSuchElementException {
        //The floatToIntBits return exactly what we are looking for: IEE 754 coding of the double
        this.write(Float.floatToIntBits(f));
    }

    /**
     * The endianness is taken care of.
     * @return a java (big endian IEEE 754) float
     */
    public float readFloat() throws NoSuchElementException {
        return Float.intBitsToFloat(this.readInt());
    }

    /**Write in a big endian, IEEE 754 way
     * @param dbl
     */
    public void write(double dbl) {
        this.write(Double.doubleToLongBits(dbl));
    }

    /**
     * The endianness is taken care of.
     * @return a java (big endian IEEE 754) double
     */
    public double readDouble() {
        return Double.longBitsToDouble(this.readLong());
    }

    /**Write a boolean, actually a byte set as 0 or 1.
     * @param f
     */
    public void write(boolean bool) {
        //The boolean isn't just a bit or byte of information in java. the offical doc says:
        //"This data type represents one bit of information, but its "size" isn't something that's precisely defined."
        if (bool) {
            buff.add((byte) 1);
        } else {
            buff.add((byte) 0);
        }
    }

    /**
     *
     * @return
     */
    public boolean readBoolean() throws NoSuchElementException {
        return (ite.next() == 1);
    }

    /**
     * Write a char in two bytes, big endian way.
     * Should be UTF-16... That's net totally sure yet.
     * @param ch
     */
    public void write(char ch) {
        //java uses utf-16... I'm nearly sure.
        //It only writes the two bytes of a character one after the other. I'm not sure this should be public, because there
        //are no equivalent in the C++ version.
        buff.add((byte) (ch >>> 8));
        buff.add((byte) ch);
    }

    /**
     * The endianness is taken care of.
     * @return java use utf-16BE chars; there are no problems with the C++ version of CERTI
     */
    public char readChar() {
        //Quite dirty... the toChars return a char[], that's why there is a [0].
        return Character.toChars((int) this.readShort())[0];

    }

    /**
     * Write a String: first an int with the size, then all the char, each one coded on two bytes in UTF-16BE.
     * @param str1 the String we want to write
     */
    public void write(String str1) {
            byte[] bytesArray = str1.getBytes();

            this.write(bytesArray.length);
            this.write(bytesArray);
        
    }

    /**
     * The endianness is taken care of.
     * @return
     */
    public String readString() {
        int length = this.readInt();

        byte[] byteArray = new byte[length];
        byteArray = this.readBytes(length);
        //try {
            //return new String(byteArray, "UTF-8");
            return new String(byteArray);
        /*} catch (UnsupportedEncodingException e) {
            e.printStackTrace();
            return null;
        }*/
    }

    /**
     * Not tested yet.
     * The buffer is reseted after returning the message.
     * @param sock the Input socket
     * @return
     * @throws IOException
     */
    public CertiMessage receive(InputStream in) throws IOException, CertiException {
        //assumeSizeFromReservedBytes
        endianness = (byte) in.read();
        int size = 0;
        //Same as readInt, but on the InputStream instead of the iterator
        if (endianness == BIG_ENDIAN) {
            for (int i = 3; i >= 0; i--) {
                size = size | ((int) in.read() & 0x000000FF) << i * 8;
            }
        } else {
            for (int i = 0; i <= 3; i++) {
                size = size | ((int) in.read() & 0x000000FF) << i * 8;
            }
        }

        for (int i = 0; i <= size - 6; i++) {
            buff.add((byte) in.read());
        }

        ite = buff.iterator();

        CertiMessage msg = CertiMessage.virtualCreate(this.readInt()); //read first byte from header and instantiate class

        /*TODO remove reflexion
        switch(CertiMessageType.reverseType.get(this.readInt()))
        {
            case CREATE_FEDERATION_EXECUTION:

        }*/

        msg.readMessage(this);
        //this.reset();
        return msg;
    }

    /**
     * Send the message throught the socket
     * @param sock
     * @throws IOException
     */
    public void send(OutputStream out) throws IOException {
        //Add size to header
        Vector<Byte> header = new Vector<Byte>();

        int size = (buff.size() + 5);
        header.add((byte) BIG_ENDIAN);
        for (int i = 3; i >= 0; i--) {
            header.add((byte) (size >>> i * 8));
        }

        //TODO zmenit nech nerobim zbytocne jak dilino
        Vector<Byte> message = new Vector<Byte>();
        message.addAll(header);
        message.addAll(buff);

        //   System.out.print("CertiMessage low level");
        byte[] byteArray = new byte[message.size()];
        for (int i = 0; i < message.size(); i++) {
            //System.out.print(message.get(i).byteValue());
            byteArray[i] = message.get(i).byteValue();
            //System.out.print(" " + message.get(i).byteValue());
            //that's where the difference between Byte and byte has an influence; we can't send Byte
        }

        
        //System.out.println(";");
        out.write(byteArray);
        this.reset();
    }

    /**
     * for tests only
     * @return the hex code of the data in the buffer
     */
    public String readAllHex() {
        String s = new String();
        for (Byte b : this.buff) {
            int i = ((int) b) & 0x000000FF;
            s = s + Integer.toHexString(i) + " ";
        }
        return s;
    }

    /**
     * for tests only
     * Must be used before reading directly from the buffer without using the receive method
     */
    public void setIterator() {
        ite = buff.iterator();
    }
}
